from mdgrad.tensor import Tensor
from .layers import Module
import numpy as np

class MSELoss(Module):
    def forward(self, x, y):
        assert x.shape == y.shape, 'input and target tensors must be the same shape'
        x = x if isinstance(x, Tensor) else Tensor(x)
        y = y if isinstance(y, Tensor) else Tensor(y)
        
        out = ((x - y) ** 2).sum() /  x.size

        return out
    
    def parameters(self):
        return []
    
class CrossEntropyLoss(Module):
    def forward(self, x, y):
        x = x if isinstance(x, Tensor) else Tensor(x)
        y = y if isinstance(y, Tensor) else Tensor(y)

        # Turn the logits into probs
        probs = x.softmax()
        samples = len(probs)
        probs_clipped = np.clip(probs.numpy(), 1e-7, 1 - 1e-7)
        if len(y.shape) == 1:
            # If y values are class indices
            confidence = probs_clipped[range(samples), y.numpy().astype(int)]
        elif len(y.shape) == 2:
            # If y values are one-hot encoded
            confidence = np.sum(probs_clipped * y.numpy(), axis=1)
        nll = -np.log(confidence)
        out = Tensor(np.mean(nll), (x,))

        def _backward():
            nonlocal y
            if type(y) != np.ndarray:
                y =y.numpy()
            if len(y.shape) == 2:
                # Turn labels to discrete values if they are
                # one-hot encoded
                y = np.argmax(y.numpy(), axis=1)
            # Gradients of inputs
            x.grad[np.arange(len(y.data)), y.astype(int)] -= 1
            x.grad += probs / len(y.data) * out.grad

        out._backward = _backward

        return out
        
    def parameters(self):
        return []


import numpy as np

class Optimizer:
    def __init__(self, parameters, lr):
        self.parameters = parameters
        self.lr = lr

    def step(self):
        raise NotImplementedError
    
    def zero_grad(self):
        for p in self.parameters:
            p.grad = np.zeros_like(p.data)

class SGD(Optimizer):
    def __init__(self, parameters, lr=0.001):
        super().__init__(parameters, lr)

    def __call__(self):
        self.step()
    
    def step(self):
        for p in self.parameters:
            p.data -= self.lr * p.grad

class Adam(Optimizer):
    def __init__(self, parameters, lr=0.001, betas=(0.9, 0.999), eps=1e-8):
        super().__init__(parameters, lr)
        self.beta1 = betas[0]
        self.beta2 = betas[1]
        self.t = 0
        self.eps = eps
        self.m = [np.zeros_like(p.data) for p in self.parameters]
        self.v = [np.zeros_like(p.data) for p in self.parameters]

    def step(self):
        self.t += 1
        for i, p in enumerate(self.parameters):
            # Update biased first moment estimate
            self.m[i] = self.beta1 * self.m[i] + (1 - self.beta1) * p.grad
            # Update biased second raw moment estimate
            self.v[i] = self.beta2 * self.v[i] + (1 - self.beta2) * (p.grad ** 2)
            # Compute bias-corrected first moment estimate
            m_hat = self.m[i] / (1 - (self.beta1 ** self.t))
            # Compute bias corrected second raw moment estimate
            v_hat = self.v[i] / (1 - (self.beta2 ** self.t))
            # Update parameters
            p.data -= self.lr * m_hat / (np.sqrt(v_hat) + self.eps)



import sys
sys.path.append('..')

import numpy as np
import mdgrad
import torch

np.random.seed(42)

def test_add(op='Addition'):
    a = np.random.rand(2, 3)
    b = np.random.rand(2, 3)

    a_m = mdgrad.tensor(a)
    b_m = mdgrad.tensor(b)
    x_m = a_m + b_m
    x_m.sum().backward()

    a_t = torch.tensor(a, requires_grad=True)
    b_t = torch.tensor(b, requires_grad=True)
    x_t = a_t + b_t
    x_t.sum().backward()

    print(op)
    print(f'mdgrad: {x_m}')
    print('\n')
    print(f'torch: {x_t}')
    print('\n\n')

    assert np.allclose(
        x_t.detach().numpy(), x_m.numpy()
    ), f'{op} results do not match'

    assert np.allclose(
        a_t.grad.detach().numpy(), a_m.grad
    ), 'Gradient results do not match'

def test_sub(op='Subtraction'):
    a = np.random.rand(2, 3)
    b = np.random.rand(2, 3)

    a_m = mdgrad.tensor(a)
    b_m = mdgrad.tensor(b)
    x_m = a_m - b_m
    x_m.sum().backward()

    a_t = torch.tensor(a, requires_grad=True)
    b_t = torch.tensor(b, requires_grad=True)
    x_t = a_t - b_t
    x_t.sum().backward()

    print(op)
    print(f'mdgrad: {x_m}')
    print('\n')
    print(f'torch: {x_t}')
    print('\n\n')

    assert np.allclose(
        x_t.detach().numpy(), x_m.numpy()
    ), f'{op} results do not match'

    assert np.allclose(
        a_t.grad.detach().numpy(), a_m.grad
    ), 'Gradient results do not match'

def test_mul(op='Multiplication'):
    a = np.random.rand(3, 2)
    b = np.random.rand(3, 2)

    a_m = mdgrad.tensor(a)
    b_m = mdgrad.tensor(b)
    x_m = a_m * b_m
    x_m.sum().backward()

    a_t = torch.tensor(a, requires_grad=True)
    b_t = torch.tensor(b, requires_grad=True)
    x_t = a_t * b_t
    x_t.sum().backward()

    print(op)
    print(f'mdgrad: {x_m}')
    print('\n')
    print(f'torch: {x_t}')
    print('\n\n')

    assert np.allclose(
        x_t.detach().numpy(), x_m.numpy()
    ), f'{op} results do not match'

    assert np.allclose(
        a_t.grad.detach().numpy(), a_m.grad
    ), 'Gradient results do not match'

def test_div(op='Division'):
    a = np.random.rand(3, 2)
    b = np.random.rand(3, 2)

    a_m = mdgrad.tensor(a)
    b_m = mdgrad.tensor(b)
    x_m = a_m / b_m
    x_m.sum().backward()

    a_t = torch.tensor(a, requires_grad=True)
    b_t = torch.tensor(b, requires_grad=True)
    x_t = a_t / b_t
    x_t.sum().backward()

    print(op)
    print(f'mdgrad: {x_m}')
    print('\n')
    print(f'torch: {x_t}')
    print('\n\n')

    assert np.allclose(
        x_t.detach().numpy(), x_m.numpy()
    ), f'{op} results do not match'

    assert np.allclose(
        a_t.grad.detach().numpy(), a_m.grad
    ), 'Gradient results do not match'

def test_sum(op='Summation'):
    a = np.random.rand(3, 2)

    a_m = mdgrad.tensor(a)
    x_m = a_m.sum()
    x_m.backward()

    a_t = torch.tensor(a, requires_grad=True)
    x_t = a_t.sum()
    x_t.backward()

    print(op)
    print(f'mdgrad: {x_m}')
    print('\n')
    print(f'torch: {x_t}')
    print('\n\n')

    assert np.allclose(
        x_t.detach().numpy(), x_m.numpy()
    ), f'{op} results do not match'

    assert np.allclose(
        a_t.grad.detach().numpy(), a_m.grad
    ), 'Gradient results do not match'

def test_reshape(op='Reshape'):
    a = np.random.rand(3, 2)

    a_m = mdgrad.tensor(a)
    x_m = a_m.reshape(2, 3)
    x_m.sum().backward()

    a_t = torch.tensor(a, requires_grad=True)
    x_t = a_t.reshape(2, 3)
    x_t.sum().backward()

    print(op)
    print(f'mdgrad: {x_m}')
    print('\n')
    print(f'torch: {x_t}')
    print('\n\n')

    assert np.allclose(
        x_t.detach().numpy(), x_m.numpy()
    ), f'{op} results do not match'

    assert np.allclose(
        a_t.grad.detach().numpy(), a_m.grad
    ), 'Gradient results do not match'

def test_matmul(op='Matmul'):
    a = np.random.rand(3, 2)
    b = np.random.rand(2, 2)

    a_m = mdgrad.tensor(a)
    b_m = mdgrad.tensor(b)
    x_m = a_m @ b_m
    x_m.sum().backward()

    a_t = torch.tensor(a, requires_grad=True)
    b_t = torch.tensor(b, requires_grad=True)
    x_t = a_t @ b_t
    x_t.sum().backward()

    print(op)
    print(f'mdgrad: {x_m}')
    print('\n')
    print(f'torch: {x_t}')
    print('\n\n')

    assert np.allclose(
        x_t.detach().numpy(), x_m.numpy()
    ), f'{op} results do not match'

    assert np.allclose(
        a_t.grad.detach().numpy(), a_m.grad
    ), 'Gradient results do not match'

def test_pow(op='Exponential'):
    a = np.random.rand(3, 2)
    b = int(np.random.randint(3, 10))

    a_m = mdgrad.tensor(a)
    x_m = a_m ** b
    x_m.sum().backward()

    a_t = torch.tensor(a, requires_grad=True)
    x_t = a_t ** b
    x_t.sum().backward()

    print(op)
    print(f'mdgrad: {x_m}')
    print('\n')
    print(f'torch: {x_t}')
    print('\n\n')

    assert np.allclose(
        x_t.detach().numpy(), x_m.numpy()
    ), f'{op} results do not match'

    assert np.allclose(
        a_t.grad.detach().numpy(), a_m.grad
    ), 'Gradient results do not match'

if __name__ == '__main__':
    test_add()
    test_sub()
    test_mul()
    test_div()
    test_sum()
    test_reshape()
    test_matmul()
    test_pow()


from mdgrad.tensor import Tensor
from .layers import Module
import numpy as np

class ReLU(Module):
    def forward(self, x):
        assert isinstance(x, Tensor), 'input must be a Tensor'
        return x.relu()
    
    def parameters(self):
        return []
    
class Sigmoid(Module):
    def forward(self, x):
        assert isinstance(x, Tensor), 'input must be a Tensor'
        return x.sigmoid()
    
    def parameters(self):
        return []
    
class Tanh(Module):
    def forward(self, x):
        assert isinstance(x, Tensor), 'input must be a Tensor'
        return x.tanh()
    
    def parameters(self):
        return []

class SoftMax(Module):
    def forward(self, x):
        assert isinstance(x, Tensor), 'input must be a Tensor'
        return x.softmax()
    
    def parameters(self):
        return []


import numpy as np
import math

class Tensor:
    def __init__(self, data, _children=()):
        self.data = data if isinstance(data, np.ndarray) else np.array(data)
        self.grad = np.zeros_like(self.data)
        # Set data to floats to division can be done
        self.data = self.data.astype(np.float32) 
        self.grad = self.grad.astype(np.float32)
        self._prev = set(_children)
        self.shape = self.data.shape
        self.size = self.data.size
        self.dtype = self.data.dtype
        self._backward = lambda: None
    
    def __add__(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other)
        out = Tensor(self.data + other.data, (self, other))

        def _backward():
            self.grad += out.grad
            other.grad += out.grad
        out._backward = _backward

        return out
    
    def __sub__(self, other):    
        other = other if isinstance(other, Tensor) else Tensor(other)
        out = Tensor(self.data - other.data, (self, other))

        def _backward():
            self.grad += out.grad
            other.grad -= out.grad
        out._backward = _backward

        return out

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            other = np.ones_like(self.data) * other
        other = other if isinstance(other, Tensor) else Tensor(other)
        out = Tensor(self.data * other.data, (self, other))

        def _backward():
            self.grad += other.data * out.grad
            other.grad += self.data * out.grad
        out._backward = _backward

        return out
    
    def __pow__(self, other):
        assert isinstance(other, (int, float)), "Exponent must be a scalar (int/float)"
        out = Tensor(self.data ** other, (self,))

        def _backward():
            self.grad += (other * self.data ** (other -1)) * out.grad
        out._backward = _backward

        return out
    
    def __matmul__(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other)
        out = Tensor(self.data @ other.data, (self, other))

        def _backward():
            self.grad += out.grad @ np.transpose(other.data)
            other.grad += np.transpose(self.data) @ out.grad
        out._backward = _backward

        return out
    
    def relu(self):
        out = self.data * (self.data > 0)
        out = Tensor(self.data * (self.data > 0), (self,))

        def _backward():
            self.grad += (self.data > 0) * out.grad
        out._backward = _backward

        return out
    
    def transpose(self, axes=None):
        out = Tensor(np.transpose(self.data, axes=axes), (self,))
        
        def _backward():
            self.grad += np.transpose(out.grad, axes=axes)
        out._backward = _backward
        
        return out
    
    def log(self):
        out = Tensor(np.log(self.data + 1e-9), (self,))

        def _backward():
            self.grad += (1/self.data) * out.grad
        out._backward = _backward

        return out
    
    def exp(self):
        exp = np.exp(self.data)
        out = Tensor(exp, (self,))
        
        def _backward():
            self.grad += exp * out.grad
        out._backward = _backward

        return out
    
    def sigmoid(self):
        def _positive(x):
            return 1 / (1 + np.exp(-x))
        def _negative(x):
            exp = np.exp(x)
            return exp / (exp + 1)
        # Numerically stable sigmoid
        positive = self.data >= 0
        negative = ~positive
        value = np.empty_like(self.data)
        value[positive] = _positive(self.data[positive])
        value[negative] = _negative(self.data[negative])
        out = Tensor(value, (self,))
        
        def _backward():
            #exp = np.exp(-self.data)
            g = value * (1 - value)
            self.grad += g * out.grad
        out._backward = _backward

        return out
    
    def tanh(self, dim=1):
        val = np.tanh(self.data)
        out = Tensor(val, (self,))

        def _backward():
            self.grad += (1 - val**2) * out.grad
        out._backward = _backward

        return out
    
    def softmax(self, dim=1):
        exps = np.exp(self.data - np.max(self.data, axis=dim, keepdims=True))
        probs = exps / np.sum(exps, axis=dim, keepdims=True)
        out = Tensor(probs, (self,))

        def _backward():
            for i, (output, grad) in enumerate(zip(out.data, out.grad)):
                output = output.reshape(-1, 1)
                jacobian = np.diagflat(output) - output @ np.transpose(output)
                g = np.transpose(jacobian @ np.transpose(grad))
                self.grad[i] += g
        out._backward = _backward

        return out
    
    def reshape(self, *new_shape):
        old_shape = self.shape
        out = Tensor(self.data.reshape(*new_shape), (self,))

        def _backward():
            self.grad += out.grad.reshape(old_shape)
        out._backward = _backward

        return out
    
    def sum(self, axis=None, keepdims=True):
        out = Tensor(np.sum(self.data, axis=axis, keepdims=keepdims), (self,))

        def _backward():
            self.grad += np.ones_like(self.data) * out.grad
        out._backward = _backward

        return out
    
    def abs(self):
        out = Tensor(np.abs(self.data), (self,))

        def _backward():
            self.grad += np.ones_like(self.data) * out.grad
        out._backward = _backward

        return out
    
    def sin(self):
        out = Tensor(np.sin(self.data), (self,))

        def _backward():
            self.grad += np.cos(self.data) * out.grad
        out._backward = _backward

        return out
    
    def cos(self):
        out = Tensor(np.cos(self.data), (self,))

        def _backward():
            self.grad += -np.sin(self.data) * out.grad
        out._backward = _backward

        return out
        
    def backward(self):
        # https://github.com/karpathy/micrograd/blob/master/micrograd/engine.py
        # topological order all of the children in the graph
        
        topo = []
        visited = set()
        def build_topo(v):
            if v not in visited:
                visited.add(v)
                for child in v._prev:
                    build_topo(child)
                topo.append(v)
        build_topo(self)

        # go one variable at a time and apply the chain rule to get its gradient
        self.grad = np.ones_like(self.data)
        for v in reversed(topo):
            v._backward() 
    
    def max(self, axis=None):
        return np.max(self.data, axis=axis)
    
    def min(self, axis=None):
        return np.min(self.data, axis=axis)
    
    def numpy(self):
        return self.data.copy()
        
    @classmethod
    def zeros(cls, shape):
        assert isinstance(shape, int) or isinstance(shape, tuple), f'shape should be int or tuple insted of {type(shape)}'
        return cls(np.zeros(shape))

    @classmethod
    def ones(cls, shape):
        assert isinstance(shape, int) or isinstance(shape, tuple), f'shape should be int or tuple insted of {type(shape)}'
        return cls(np.ones(shape))
    
    @classmethod
    def normal(cls, mean=0.0, std=1.0, shape=None):
        assert isinstance(shape, int) or isinstance(shape, tuple), f'shape should be int or tuple insted of {type(shape)}'
        return cls(np.random.normal(mean, std, shape))
    
    @classmethod
    def randn(cls, *args):
        return cls(np.random.randn(*args))
    
    @classmethod
    def eye(cls, N, M=None):
        return cls(np.eye(N, M))

    def __neg__(self):
        return self * -1
    
    def __radd__(self, other):
        return self + other
    
    def __rsub__(self, other):
        return other - self
    
    def __rmul__(self, other):
        return self * other
    
    def __truediv__(self, other):
        return self * other**-1
    
    def __rtruediv__(self, other):
        return other * self**-1
    
    def __repr__(self):
        return f'Tensor(data={self.data}, dtype={self.data.dtype})'
    
    def __len__(self):
        return len(self.data)
    
    def __iter__(self):
        # Track the current element in the iterable
        self.current = 0
        return self
    
    def __next__(self):
        if self.current >= len(self.data):
            raise StopIteration
        current = self.data[self.current]
        self.current += 1
        return current
    
    def __getitem__(self, key):
        return self.data[key]
    
    def __setitem__(self, key, value):
        self.data[key] = value

def sum(x, axis=None, keepdims=True):
    x = x if isinstance(x, Tensor) else Tensor(x)
    return x.sum(axis=axis, keepdims=keepdims)

def mean(x, axis=None, keepdims=False):
    x = x if isinstance(x, Tensor) else Tensor(x)
    out = x.sum(axis=axis, keepdims=keepdims) 
    return out * math.prod(out.shape) / math.prod(x.shape)

def exp(x):
    x = x if isinstance(x, Tensor) else Tensor(x)
    return x.exp()

def log(x):
    x = x if isinstance(x, Tensor) else Tensor(x)
    return x.log()

def sin(x):
    x = x if isinstance(x, Tensor) else Tensor(x)
    return x.sin()

def cos(x):
    x = x if isinstance(x, Tensor) else Tensor(x)
    return x.cos()

def max(x, axis=None):
    x = x if isinstance(x, Tensor) else Tensor(x)
    return x.max(axis=axis)

def min(x, axis=None):
    x = x if isinstance(x, Tensor) else Tensor(x)
    return x.min(axis=axis)

def argmax(x, axis=None):
    x = x if isinstance(x, Tensor) else Tensor(x)
    return Tensor(np.argmax(x.data, axis=axis))

def argmin(x, axis=None):
    x = x if isinstance(x, Tensor) else Tensor(x)
    return Tensor(np.argmin(x.data, axis=axis))

def transpose(x, axes=None):
    x = x if isinstance(x, Tensor) else Tensor(x)
    return x.transpose(axes=axes)

def ones(shape):
    return Tensor.ones(shape=shape)

def zeros(shape):
    return Tensor.zeros(shape=shape)

def eye(N, M=None):
    return Tensor.eye(N=N, M=M)

def randn(*shape):
    return Tensor.randn(*shape)

def tensor(data):
    return Tensor(data)


from setuptools import setup, find_packages
import os

VERSION = '0.2.1'
DESCRIPTION = 'Tensor-based autdiff engine and neural network API'
LONG_DESCRIPTION = """

# mdgrad

A small autograd engine that implements backpropagation (reverse-mode autodiff). Heavily inspired by karpathy's [micrograd](https://github.com/karpathy/micrograd/tree/master), and extended to support operations on tensors instead of scalars. Includes a small neural network api for building and training neural networks. Has a PyTorch-like API.

Hopefully useful as an educational resource.

## Installation

``` bash
pip install mdgrad
```

## Example Usage

A dumb example showing supported operations

```python

import mdgrad
import mdgrad.nn as nn

a = 3 * mdgrad.randn(3, 2)
b = mdgrad.ones(shape=(2, 2))
c = a @ b
d = c * 3 / 2
e = d ** 2
f = e.sum()
print(f.data) 
f.backward()
print(a.grad) 
```

An example showing how to define and run a neural network. See the files in `examples/` for more details on building and training models.

```python

import mdgrad
import mdgrad.nn as nn

# Define the model and loss function
model = nn.Sequential([
    nn.Linear(2, 20),
    nn.ReLU(),
    nn.Linear(20, 50), 
    nn.ReLU(),
    nn.Linear(50, 15),
    nn.ReLU(),
    nn.Linear(15, 1),
    nn.Sigmoid()
])
loss_fn = nn.MSELoss()

# Create dummy data
X = mdgrad.randn(100, 2)
target = mdgrad.randn(100, 1)

# Compute output and loss
out = model(X)
loss = loss_fn(out, target)

# Compute gradients of parameters
loss.backward()
```

                 """

# Setting up
setup(
    name="mdgrad",
    version=VERSION,
    author="Jordan Madden",
    author_email="<jordanmadden285@gmail.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=['numpy'],
    keywords=['python', 'tensors', 'neural networks', 'automatic differentiation'],
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
    python_requires='>=3.8'
)


from mdgrad.tensor import Tensor
import numpy as np

class Module:
    def __call__(self, *x):
        out = self.forward(*x)
        return out
    
    def forward(self, x):
        raise NotImplementedError

    def parameters(self):
        params = []
        for key, value in self.__dict__.items():
            if isinstance(value, Tensor):
                params.append(value)
            if isinstance(value, Module):
                params.extend(value.parameters())
        return list(set(params))
    
class Linear(Module):
    def __init__(self, in_features, out_features, bias=True):
        self.bias = bias
        self.w = Tensor.randn(in_features, out_features) / np.sqrt(in_features + out_features)
        self.b = Tensor.zeros((1, out_features)) if self.bias else None

    def forward(self, x):
        x = x if isinstance(x, Tensor) else Tensor(x)
        
        out = x @ self.w + self.b if self.bias else x @ self.w
        # Add previous tensors to computation graph
        out._prev = set((x,))

        def _backward():
            # Gradients of parameters
            self.w.grad += x.transpose() @ out.grad
            if self.bias:
                self.b.grad += out.grad.sum(0, keepdims=True)
            # Gradients of inputs
            x.grad += out.grad @ self.w.transpose()
        out._backward = _backward

        return out

    def parameters(self):
        if self.bias:
            return [self.w, self.b]
        else:
            return [self.w]

class Sequential(Module):
    def __init__(self, layers=[]):
        super().__init__()
        assert isinstance(layers, (list, tuple)), 'input must be a list or tuple'
        self.layers = layers

    def forward(self, x):
        for layer in self.layers:
            x = layer(x)
        return x
    
    def parameters(self):
        params = []
        for layer in self.layers:
            params.extend(layer.parameters())
        return params